"""

  FITS document visualizer.

  @author Juan Pedro Bolivar Puente
  @date   April 2011

"""

from itertools import repeat
import logging

import gtk
import gobject
import pyfits as fits

import numpy
from matplotlib.figure import Figure
from matplotlib import cm

# Descomentar para elegir GTK, GTKAgg o GTKCairo
#
# from matplotlib.backends.backend_gtk \
#     import FigureCanvasGTK as FigureCanvas
# from matplotlib.backends.backend_gtkagg \
#     import FigureCanvasGTKAgg as FigureCanvas
from matplotlib.backends.backend_gtkcairo \
    import FigureCanvasGTKCairo as FigureCanvas

import img_scale
import util


GUI_PRIMARY = 'gui/primary.glade'
GUI_TABLE   = 'gui/table.glade'

class FitsError (Exception): pass
class FitsDimensionsError (FitsError): pass
class FitsDataError (FitsError): pass

class FitsImagePage (object):

    curr_layer   = -1
    must_combine = False

    def __init__ (self, hdu = None, *a, **k):
        super (FitsImagePage, self).__init__ (*a, **k)

        builder = gtk.Builder ()
        builder.add_from_file (GUI_PRIMARY)
        builder.connect_signals (self)

        #page       = builder.get_object ('primary-frame')
        #container  = builder.get_object ('primary-frame-align')
        page       = builder.get_object ('image-container')
        container  = page
        toolbar    = builder.get_object ('primary-toolbar')
        layer      = builder.get_object ('combo-layer')
        combine    = builder.get_object ('tool-combine')
        
        figure = Figure (figsize = (1,1))
        axes   = figure.add_subplot (111)
        container.add (FigureCanvas (figure))
        container.show_all ()
        
        cell = gtk.CellRendererText ()
        layer.pack_start (cell, True)
        layer.add_attribute (cell, 'text', 0)  

        self.layer   = layer
        self.figure  = figure
        self.axes    = axes
        self.hdu     = hdu
        
        self.label   = gtk.Label ('Image: ' + hdu.name)
        self.page    = page
        self.toolbar = toolbar
        
        self.update_image ()
        
        if hdu.data is None or hdu.data.ndim == 2:
            combine.set_sensitive (False)
            layer.set_sensitive (False)
            layer.append_text ('No layers')
            
        elif hdu.data.ndim == 3:
            self.curr_layer = 0
            for x in xrange (len (hdu.data)):
                layer.append_text ('Layer: ' + str (x))
            layer.set_active (0)
            
        else:
            raise FitsDimensionsError, 'Wrong number of dimensions.'
    
    def handle_toggle_combine (self, obj):
        if obj.get_active () and (
            self.hdu.data is None or
            not (self.hdu.data.ndim == 3) or
            not (len (self.hdu.data) == 3)):
            util.show_error_dialog (
                'Wrong operation',
                'Layers can be combined only if there are exactly 3 layers.')
            obj.set_active (False)
        else:
            self.must_combine = obj.get_active ()
            self.layer.set_sensitive (not self.must_combine)

        self.update_image ()
        
    def handle_change_layer (self, obj):
        self.curr_layer = obj.get_active ()
        self.update_image ()

    def update_image (self):
        data = self.hdu.data

        if data is None:
            logging.warning ('Empty Primary HDU.')
            return
        elif data.ndim == 3:
            w, h = data [0].shape
            if self.must_combine:
                img = numpy.zeros ((w, h, 3), dtype = float)
                for i in xrange (3):
                    img [:, :, i] = img_scale.doc (data [i], 0, 256)
            else:
                img = img_scale.sqrt (data [self.curr_layer], 0, 256)

        else:
            w, h = data.shape
            img = img_scale.sqrt (data, 0, 255)

        self.axes.imshow (img, cmap=cm.gray)
        self.figure.canvas.draw ()


format_to_gobj = {
    'L' : gobject.TYPE_BOOLEAN,
    'X' : gobject.TYPE_BOOLEAN,
    'B' : gobject.TYPE_UINT,
    'I' : gobject.TYPE_INT,
    'J' : gobject.TYPE_INT,
    'K' : gobject.TYPE_INT,
    'A' : gobject.TYPE_CHAR,
    'E' : gobject.TYPE_FLOAT,
    'D' : gobject.TYPE_DOUBLE
    }

def get_column_gtype (col):
    try:
        return format_to_gobj [col.format [0]]
    except KeyError:
        return gobject.TYPE_STRING
        
class FitsTablePage (object):

    def __init__ (self, hdu = None, *a, **k):
        super (FitsTablePage, self).__init__ (*a, **k)

        builder = gtk.Builder ()
        builder.add_from_file (GUI_TABLE)
        builder.connect_signals (self)

        #page  = builder.get_object ('table-frame')
        page  = builder.get_object ('table-box')
        view  = builder.get_object ('table-view')
        
        store = gtk.ListStore (* map (get_column_gtype, hdu.columns))
        view.set_model (store)

        for i, col in enumerate (hdu.columns):
            view.append_column (gtk.TreeViewColumn (
                col.name, gtk.CellRendererText (), text = i))

        for row in hdu.data:
            store.append (row)
        
        self.hdu     = hdu
        self.label   = gtk.Label ('Table: ' + hdu.name)
        self.page    = page
        self.toolbar = None


page_types = {
    fits.PrimaryHDU   : FitsImagePage,
    fits.ImageHDU     : FitsImagePage,
    fits.TableHDU     : FitsTablePage,
    fits.GroupsHDU    : FitsTablePage,
    fits.BinTableHDU  : FitsTablePage
    }

class FitsDocument (object):

    def __init__ (self, filename = '', *a, **k):    
        super (FitsDocument, self).__init__ (*a, **k)
        
        f = fits.open (filename)

        self.hdus = [ page_types [x.__class__] (x) for x in f ]
        self._file = f
        
    def __enter__ (self):
        pass
    
    def __leave__ (self):
        self.close ()

    def __del__ (self):
        self.close ()
        
    def close (self):
        self._file.close ()
